/****************************************************************************
**
** Copyright (C) 2024 MikroElektronika d.o.o.
** Contact: https://www.mikroe.com/contact
**
** This file is part of the mikroSDK package
**
** Commercial License Usage
**
** Licensees holding valid commercial NECTO compilers AI licenses may use this
** file in accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The MikroElektronika Company.
** For licensing terms and conditions see
** https://www.mikroe.com/legal/software-license-agreement.
** For further information use the contact form at
** https://www.mikroe.com/contact.
**
**
** GNU Lesser General Public License Usage
**
** Alternatively, this file may be used for
** non-commercial projects under the terms of the GNU Lesser
** General Public License version 3 as published by the Free Software
** Foundation: https://www.gnu.org/licenses/lgpl-3.0.html.
**
** The above copyright notice and this permission notice shall be
** included in all copies or substantial portions of the Software.
**
** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
** OF MERCHANTABILITY, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
** TO THE WARRANTIES FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
** DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
** OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**
****************************************************************************/

/**
 * \file gl_image.h
 * \brief API for image drawing. Here can be found functions for getting information about image or drawing it.
 */
#ifndef _GL_IMAGE_H_
#define _GL_IMAGE_H_

#include "gl_types.h"
#include "generic_pointer.h"
#include "gl_shapes.h"

/** \addtogroup apigroup API
 *  \brief API
 *  @{
 */

/**
 * \addtogroup glgroup Graphic Library
 * \brief Graphic Library
 *  @{
 */

#ifdef __cplusplus
extern "C"{
#endif

/**
 * \brief Draw image on display.
 *
 * \details Draw image on display for all supported formats. Function specialized for drawing exact image format, can be redefined by user.
 * For that option look at gl_image_format_handlers.h .
 *
 * \param[in] dest  Rectangle that represents destination where picture wil be drawn. See \ref gl_rectangle_t structure definition for detailed explanation.
 * \param[in] src Rectangle that represents part of image that will be draw into destination, et. \p dest rectangle. See \ref gl_rectangle_t structure definition for detailed explanation.
 * \param[in] image Pointer to image.
 *
 * \return Returns zero if image is successfuly drawn, otherwise a number greater then zero iz returned.
 *
 * \pre Before drawing driver must be set by #gl_set_driver.
 *
 * \note Image must be generated by NectoStudio's resource generator.
 *
 *  \sa #gl_draw_jpeg_image, #gl_draw_bitmap_16bpp, #gl_draw_bitmap_8bpp, #gl_draw_bitmap_4bpp, #gl_draw_bitmap_1bpp
 */
int gl_draw_image(gl_rectangle_t *dest, gl_rectangle_t *src, const uint8_t * __generic_ptr image);

/**
 * \brief Gives width of the image.
 *
 * \details The information is read from the image.
 *
 * \param[in] image Pointer to image.
 *
 * \return If the image is not null returns width of the image, otherwise returns zero.
 *
 * \note Image must be generated by NectoStudio's resource generator.
 */
uint16_t gl_image_width(const uint8_t * __generic_ptr image);

/**
 * \brief Gives height of the image.
 *
 * \details The information is read from the image.
 *
 * \param[in] image Pointer to image.
 *
 * \return If the image is not null returns height of the image, otherwise returns zero.
 *
 * \note Image must be generated by NectoStudio's resource generator.
 */
uint16_t gl_image_height(const uint8_t * __generic_ptr image);


/**
 * \brief Gives format of the image.
 *
 * \details The information is read from the image.
 *
 * \param[in] image Pointer to image.
 *
 * \return If the image is not null returns format of image, otherwise returns zero. See \ref gl_image_format_t structure definition for detailed explanation.
 *
 * \note Image must be generated by NectoStudio's resource generator.
 */
gl_image_format_t gl_image_format(const uint8_t * __generic_ptr image);

#ifdef __cplusplus
} // extern "C"
#endif

/** @} */ // glgroup
/** @} */ // apigroup

#endif // _GL_IMAGE_H_
// ------------------------------------------------------------------------- END
